import logging
import random
import time
import json
import pandas as pd
import numpy as np
import scipy
import scipy.stats
from sklearn.metrics import mean_squared_error
import pywt
import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import plotly.graph_objs as go
import plotly.express as px
from hurst import compute_Hc
from RegscorePy.aic import aic
from RegscorePy.bic import bic
from pcap2csv.pcap2csvutils import Pcap2csvutils
from .tmmetricsutilsusess import TmMetricsUtilsUsess
from .const import conn_pkts_reqd_metrics, pkt_graph_metrics, pdu_graph_metrics
from .const import per_pkt_metrics, metric_units, time_resolution

logger = logging.getLogger(__name__)

plot_symbols = [205, 135, 133, 300]


class DashBoard():
    def __init__(self, pkt_pdus_dfs):
        self.width = 11 if len(pkt_pdus_dfs) > 2 else 10
        self.col_width = 3 if len(pkt_pdus_dfs) > 3 else 4 
        self.col_width = self.col_width if len(pkt_pdus_dfs) == 3 else 6 
        self.col_width = 12 if len(pkt_pdus_dfs) == 1 else self.col_width 
        st = [dbc.themes.BOOTSTRAP]
        self.app = dash.Dash(external_stylesheets=st)
        self.col_style = {
            'borderWidth': '1px',
            'borderStyle': 'solid',
            'borderRadius': '20px',
            'borderColor': 'grey',
            'padding': '20px 20px 20px 20px'
        }

        self.input_state = []
        self.outputs = []

        self.xy_vals = None
        self.start_time = 0

    def get_applications(self, pkt_df):
        apps = [{'label':'all', 'value':'all'}]
        apps.extend([{'label':x, 'value':x} for x in pkt_df.app_idx.unique().tolist()])
        return apps
    
    def get_input_col(self, idx, pkts_df):  
        if len(pkts_df.index) == 0:
            dur, convs, convs_val, endpts, endpts_val, apps, apps_val = [None, None, None, None, None, None, None]
        else:
            dur = pkts_df.iloc[-1].time - pkts_df.iloc[0].time
            convs = self.get_conversations(pkts_df)
            convs_val = convs[0]['value'] if len(convs)>0 else ''
            endpts = self.get_endpoints(pkts_df)
            endpts_val = convs[0]['value'] if len(endpts)>0 else ''
            apps = self.get_applications(pkts_df)
            apps_val = apps[0]['value'] if len(apps)>0 else ''
        return dbc.Col(
                    html.Div([
                        html.H3('{} {}:'.format(self.analysis_type, idx)),
                        html.P('{} {}, in {} seconds'.format(len(pkts_df), self.analysis_type, dur)),
                        dbc.FormGroup([
                            dbc.Label('Filter String', html_for='filter-textbox-{}'.format(idx)), 
                            dbc.Input(
                                type='text',
                                id='filter-textbox-{}'.format(idx),
                                placeholder='',
                            ),
                            dbc.FormText(
                                'leave empty to select all packets',
                                color='secondary',
                            )
                        ]),
                        dbc.FormGroup([
                            dbc.Label('Conversations', html_for='conv-dropdown-{}'.format(idx)), 
                            dcc.Dropdown(
                                id='conv-dropdown-{}'.format(idx),
                                options=convs,
                                value=convs_val
                            ),
                            dbc.FormText(
                                'select a specific layer 2, 3 or 4 conversation',
                                color='secondary',
                            )
                        ]),
                        dbc.FormGroup([
                            dbc.Label('Endpoints', html_for='endpt-dropdown-{}'.format(idx)), 
                            dcc.Dropdown(
                                id='endpt-dropdown-{}'.format(idx),
                                options=endpts,
                                value=endpts_val
                            ),
                            dbc.FormText(
                                'select a specific layer 2, 3 or 4 endpoint',
                                color='secondary',
                            )
                        ]),
                        dbc.FormGroup([
                            dbc.Label('Apps', html_for='app-dropdown-{}'.format(idx)), 
                            dcc.Dropdown(
                                id='app-dropdown-{}'.format(idx),
                                options=apps,
                                value=apps_val
                            ),
                            dbc.FormText(
                                'select a specific Application',
                                color='secondary',
                            )
                        ])
                    ],
                    style=self.col_style
                    )
                ,
                width=self.col_width
            )

    def get_input_row(self):
        cols = []
        for i, pkt_pdu_df in enumerate(self.pkt_pdus_dfs):
            cols.append(self.get_input_col(i, pkt_pdu_df))
        return dbc.Row(cols)

    def get_metric_input_row(self):
        m_style = dict(self.col_style, display='flex', justifyContent='center') 
        pmetrics = [{'label':x, 'value':x} for x in self.gmetrics]
        pmetrics_val = 'select_metric'
        return  dbc.Row([
                dbc.Col(
                    html.Div(children = [
                        html.H3('Select Metric: '),
                        dcc.Dropdown(
                            id='metric-dropdown',
                            options=pmetrics,
                            value=pmetrics_val,
                            style={'width':'70%'}),
                        dbc.Button('    GO!     ', id="go-button", className="mr-1", style={'width':'10%'}),
                    ],
                    style=m_style
                    )
                ,
                width=12
            )
        ])

    def get_analysis_col(self, idx):
        return  dbc.Col(
                    html.Div([
                        html.P('_ {}, in _ seconds'.format(self.analysis_type), id='summary_{}'.format(idx)),

                        dcc.Graph(id='time_series_{}'.format(idx)),

                        dbc.Row([
                            dbc.Col(
                                html.Div('stats...', id='stats_{}'.format(idx)),
                                width=4
                            ),
                            dbc.Col([
                                html.H5('Histograms'), 
                                dcc.Graph(id='hist_{}'.format(idx)), 
                                dbc.FormGroup([
                                    dbc.Label('Enter Bin Size', html_for='hist-binsize-input-{}'.format(idx)), 
                                    dbc.Input(
                                        type='text',
                                        id='hist-binsize-input-{}'.format(idx),
                                        placeholder='1000'
                                    )
                                    ])
                                ],
                                width=8
                            )
                        ]
                        )
                    ],
                    style=self.col_style
                    ),
                width=self.col_width
            )

    def get_analysis_row(self):
        cols = []
        for i in range(len(self.pkt_pdus_dfs)):
            cols.append(self.get_analysis_col(i))
        return dbc.Row(cols)

    def get_compare_row(self):
        return  dbc.Row([
                dbc.Col(
                    html.Div(
                    children = [
                        html.H5('CDF Plots'),
                        dcc.Graph(id='cdfs'),
                    ],
                    style = self.col_style
                    ),
                    width=4
                ),
                dbc.Col(
                    html.Div(
                    children = [
                        html.H5('Q-Q Plots'),
                        dcc.Graph(id='qqplots'),
                    ],
                    style = self.col_style
                    ),
                    width=4
                ),
                dbc.Col(
                    html.Div(
                    children = [
                        html.H5('comparison Metrics'),
                        html.Div('...', id='comp_metrics')
                    ],
                    style = self.col_style
                    ),
                    width=4
                ),
            ])

    def get_compare_row2(self):
        return  dbc.Row([
            dbc.Col(
                html.Div(
                children = [
                    html.H5('Box Plots'),
                    dcc.Graph(id='boxplo'),
                ],
                style = self.col_style
                ),
                width=4
            ),
            dbc.Col(
                html.Div(
                children = [
                    html.H5('Wavelet Scaling Plots'),
                    dcc.Graph(id='waveletplots'),
                ],
                style = self.col_style
                ),
                width=4
            ),
            dbc.Col(
                html.Div(
                children = [
                    html.H5('...'),
                    html.Div('...', id='plot6')
                ],
                style = self.col_style
                ),
                width=4
            ),
        ])

    def get_timeseries(self, x, y, metric):
        if metric == 'connection_durations':
            return self.get_conn_dur_graph(x, y, metric)
        title = metric
        m =len(x)
        if m > 2000000:
            x = []
            y = []
            title = title + '- dataset too long (' + str(m) + ') - skipping' 
        gtype = go.Bar if metric in per_pkt_metrics else go.Scatter
        logger.debug('ts type = %s', str(gtype))
        if gtype == go.bar:
            ts = {'data':[go.Scattergl(x=x, y=y, width=2, 
                                       marker = {symbol: 'square-dot', color: '#cc000', size: 12 },)],
                  'layout':go.Layout(title=title,
                                     xaxis={'title': 'time(s)','showgrid': False}, 
                                     yaxis={'title': metric,'showgrid': False})
                 }
        else:
            ts = {'data':[go.Scattergl(x=x, y=y, mode='lines')],
                  'layout':go.Layout(title=title,
                                     xaxis={'title': 'time(s)','showgrid': False}, 
                                     yaxis={'title': metric,'showgrid': False})
                 }
        return ts


    def get_conn_dur_graph(self, x, y, metric):
        #connection start time = x, duration = y 
        assert (len(x) == len(y)), 'length of x must be equal to length of y'
        title = metric
        m =len(x)
        if m > 3000:
            x = []
            y = []
            title = title + '- dataset too long (' + str(m) + ') - skipping' 

        data = []
        for i in range(len(x)):
            data.append(
                go.Scatter(
                    x=[x[i], x[i]+y[i]], 
                    y=[i,i],
                    name='conn' + str(i),
                    mode='lines',
                    line={'width':4}
                )
            )
        cdfg = {
            'data':data,
            'layout':go.Layout(
                title=metric + ' Connection Durations',
                xaxis =  {          
                    'title': 'time',                           
                    'showgrid': False
                            },
                yaxis = {        
                    'title': 'Connection Durations',                    
                    'showgrid': False
                        }
            )
        }
        return cdfg

    def get_stats(self, x, y, metric):
        logger.debug('obtaining metric statistics')
        ret = TmMetricsUtilsUsess().get_stats(x, y, metric)

        #logger.debug ('@@@@@@ stats',ret)
        r = 'METRIC STATISTICS: '
        for name, val  in ret.items():
            r = r +  '   ||   ' + name + ': ' + str(val)
        return self.generate_table(r)

    def get_bins(self, y, metric, wid=None):
        st = min(y) if len(y) > 0 else 0
        en = max(y) if len(y) > 0 else 0
        wi = (en - st)/20
        wi = 1 if (wi < 1 and metric != inter_packet_time) else wi
        wi = 1000 if wid == None else wid
        #wi = wi if wid == None else wid
        bins = dict(start=st, end=en, size=wi)
        return bins, wi

    def get_hist(self, x, y, metric, binsize = None): 
        #get_hist must return [fig, binsize]
        bins, wi = self.get_bins(y, metric, wid=binsize)
        his = {
            'data':[go.Histogram(x=y, xbins=bins,
                                 autobinx=False, opacity=0.9,
                                 marker={"line": {"color": "#000000", 
                                                  "width": 0.5}})],
            'layout':go.Layout(
                title = metric + '_hist_',
                yaxis = {'title': 'Frequency','showgrid': False},
                xaxis = {'title': metric, 'showgrid': False}
            )
        }
        return [his, wi]


    def get_cdfs(self, metric):
        data = []
        for idx, [x,y] in self.xy_vals.items():
            symb = plot_symbols[idx % len(plot_symbols)]
            dd = [float(m) for m in y]
            if len(dd) > 0:
                dd.append(dd[-1])
            cdf_x = np.sort(dd)
            cdf_y = 1. * np.arange(len(dd))/(len(dd) - 1)
            data.append(
                go.Scatter(
                    x=cdf_x, 
                    y=cdf_y,
                    name='pcap_' + str(idx),
                    mode='lines+markers',
                    marker_symbol=symb,
                    marker_size=10
                )
            )
        cdfg = {
            'data':data,
            'layout':go.Layout(
                title=metric + ' CDFs',
                xaxis =  {          
                    'title': metric,                           
                    'showgrid': False
                            },
                yaxis = {        
                    'title': 'F(x)',                    
                    'showgrid': False
                        }
            )
        }
        return cdfg

    def get_qqplots(self, metric):

        '''
        if len(self.pkt_pdus_dfs) <= 1:
            return {
                'data':[],
                'layout':go.Layout(
                    title="QQ plots",
                    xaxis =  {                                     
                        'showgrid': False
                                },
                    yaxis = {                              
                        'showgrid': False
                            }
                )}'''
        data = []
        quantiles = np.linspace(start=0, stop=1, num=50)
        
        for i in range(1, len(self.pkt_pdus_dfs)):
            symb = plot_symbols[i % len(plot_symbols)]
            if len(self.xy_vals[0][1]) < 1 or len(self.xy_vals[i][1]) < 1: 
                #logger.debug('****gere****')
                data.append(
                    go.Scatter(
                        x=[], 
                        y=[],
                        mode='markers',
                        name='pcap1 vs pcap' + str(i),
                        marker_symbol=symb,
                        marker_size=15
                    )
                )
                continue

            x = self.xy_vals[0][1]
            y = self.xy_vals[i][1]
            #print self.xy_vals[0]
            #print self.xy_vals[i]

            x_quantiles = np.quantile(x, quantiles, interpolation='linear')
            y_quantiles = np.quantile(y, quantiles, interpolation='linear')
            data.append(
                go.Scatter(
                    x=x_quantiles, 
                    y=y_quantiles,
                    mode='markers',
                    name='pcap1 vs pcap' + str(i),
                    marker_symbol=symb,
                    marker_size=15
                )
            )
        qqplt = {
            'data':data,
            'layout':go.Layout(
                title=metric + " 2 sample QQ plots",
                xaxis =  {          
                    'title': metric,                           
                    'showgrid': False
                            },
                yaxis = {        
                    'title': metric,                    
                    'showgrid': False
                        }
            )
        }
        return qqplt

    def get_wavlet_scaling_plots(self, metric):
        logger.debug('obtaining wavelet stats')
        if len(self.pkt_pdus_dfs) <= 1:# or metric in per_pkt_metrics:
            return {
                'data':[],
                'layout':go.Layout(
                    title="Wavelet Scaling Energy Plots",
                    xaxis =  {      
                        'title': 'Time Scale j',
                        'showgrid': False
                                },
                    yaxis = {
                        'title': 'Log2(Energy(j))',
                        'showgrid': False
                            }
                )}
        data = []
        
        for i in range(len(self.pkt_pdus_dfs)):
            symb = plot_symbols[i % len(plot_symbols)]

            if len(self.xy_vals[i][1]) < 1:
                continue
            y = self.xy_vals[i][1]
            coeffs = pywt.wavedec(y, 'haar')
            #print 'coeffs', coeffs
            
            y_wavelet = [np.sqrt(np.sum(np.array(coeffs[-k]) ** 2)) / len(coeffs[-k]) for k in range(len(coeffs),0,-1)]
            data.append(
                go.Scatter(
                    x=range(1,len(y_wavelet)+1), 
                    y=y_wavelet,
                    mode='lines+markers',
                    name=self.analysis_type + str(i),
                    marker_symbol=symb,
                    marker_size=15
                )
            )
        wvplt = {
            'data':data,
            'layout':go.Layout(
                title=metric + " Wavelet plots",
                xaxis =  {                  
                    'title': 'Time Scale j',                   
                    'showgrid': False
                            },
                yaxis = {          
                        'title': 'Log2(Energy(j))'                    
                        }
            )
        }
        return wvplt

    def get_boxplots(self, metric):
        data = []
        
        for idx, [x,y] in self.xy_vals.items():
            data.append( go.Box(y=np.array(y)))

        bplot = {
            'data':data,
            'layout':go.Layout(
                title=metric + " Box Plots",
                    xaxis =  {                                     
                        'showgrid': False
                                },
                    yaxis = {                              
                        'showgrid': False
                            }
            )
        }
        return bplot

    def get_comp_metrics(self, metric):
        logger.debug('obtaining comparison stats')
        tmmu = TmMetricsUtilsUsess()
        if len(self.pkt_pdus_dfs) <= 1:
            return []
        pairsss = []
        for i in range(len(self.pkt_pdus_dfs)-1):
            for j in range(i+1, len(self.pkt_pdus_dfs)):
                pairsss.append((i,j))
        r2 = 'comparison metrics '
        r = []
        
        for i,j in pairsss:
            logger.debug('before getting compare metrics ' + str(i) + ' vs ' + str(j) + ': ' + str(time.time() - self.start_time))
            #print(len(self.xy_vals[i][1]), len(self.xy_vals[j][1]))
            ret = tmmu.get_comp_metrics(self.xy_vals[i][1], self.xy_vals[j][1], metric)
            name = self.analysis_type + ' ' + str(i) + '-vs-' + str(j)
            
            r2 = r2 + '||' + name + ': '
            for compmet, val in ret.items():
                r2 = r2 + '||' + compmet + ':' + val

        return self.generate_table(r2)

    def generate_table(self, r2):
        #print r2
        aa = r2.split('||')
        m = aa.pop(0)
        if len(aa) <= 1:
            return r2
        return html.Table(
            # Header
            [html.Tr([html.Th(m,colSpan='2')])] +

            # Body
            [html.Tr([ html.Td(aa[i].split(':')[0]),  html.Td(aa[i].split(':')[1]) ]) for i in range(len(aa))]
        )

    def get_layout(self):
        return html.Div(
            [
                dbc.Row([
                    dbc.Col(html.Div(),width=(12-self.width)/2),
                    dbc.Col(
                        html.Div(
                            className='one column',
                            children=[
                                html.H1('Traffic Metrics'),

                                self.get_input_row(),

                                html.P(),

                                self.get_metric_input_row(),

                                html.P(),

                                self.get_analysis_row(),

                                html.P(),

                                self.get_compare_row(),

                                html.P(),

                                self.get_compare_row2(),

                            ]
                        ),
                        width=self.width
                    ),

                    dbc.Col(
                        html.Div(),
                        width=(12-self.width)/2
                    )
                ]
                )
            ]
        )


        def store_state_components(self):
            for idx in range(len(self.pkt_dfs)):
                self.input_state.append(dash.dependencies.State('filter-textbox-{}'.format(idx), 'value'))
                self.input_state.append(dash.dependencies.State('conv-dropdown-{}'.format(idx), 'value'))
                self.input_state.append(dash.dependencies.State('endpt-dropdown-{}'.format(idx), 'value'))
                self.input_state.append(dash.dependencies.State('app-dropdown-{}'.format(idx), 'value'))

                self.outputs.append(dash.dependencies.Output('summary_{}'.format(idx), 'children'))
                self.outputs.append(dash.dependencies.Output('time_series_{}'.format(idx), 'figure'))
                self.outputs.append(dash.dependencies.Output('stats_{}'.format(idx), 'children'))
                self.outputs.append(dash.dependencies.Output('hist_{}'.format(idx), 'figure'))
                self.outputs.append(dash.dependencies.Output('hist-binsize-input-{}'.format(idx), 'value'))


                self.bin_inputs.append(dash.dependencies.Input('hist-binsize-input-{}'.format(idx), 'value'))

                self.bin_outputs.append(dash.dependencies.Output('hist_{}'.format(idx), 'figure'))

            self.input_state.append(dash.dependencies.State('metric-dropdown', 'value'))
            self.outputs.append(dash.dependencies.Output('cdfs', 'figure'))
            self.outputs.append(dash.dependencies.Output('qqplots', 'figure'))
            self.outputs.append(dash.dependencies.Output('comp_metrics', 'children'))
            self.outputs.append(dash.dependencies.Output('boxplo', 'figure'))
            self.outputs.append(dash.dependencies.Output('waveletplots', 'figure'))

    def append_dependencies(self):
        for idx in range(len(self.pkt_pdus_dfs)):
            self.input_state.append(dash.dependencies.State('filter-textbox-{}'.format(idx), 'value'))
            self.input_state.append(dash.dependencies.State('conv-dropdown-{}'.format(idx), 'value'))
            self.input_state.append(dash.dependencies.State('endpt-dropdown-{}'.format(idx), 'value'))
            self.input_state.append(dash.dependencies.State('app-dropdown-{}'.format(idx), 'value'))
            self.input_state.append(dash.dependencies.State('hist-binsize-input-{}'.format(idx), 'value'))

            self.outputs.append(dash.dependencies.Output('summary_{}'.format(idx), 'children'))
            self.outputs.append(dash.dependencies.Output('time_series_{}'.format(idx), 'figure'))
            self.outputs.append(dash.dependencies.Output('stats_{}'.format(idx), 'children'))
            self.outputs.append(dash.dependencies.Output('hist_{}'.format(idx), 'figure'))
            self.outputs.append(dash.dependencies.Output('hist-binsize-input-{}'.format(idx), 'value'))

        self.input_state.append(dash.dependencies.State('metric-dropdown', 'value'))
        self.outputs.append(dash.dependencies.Output('cdfs', 'figure'))
        self.outputs.append(dash.dependencies.Output('qqplots', 'figure'))
        self.outputs.append(dash.dependencies.Output('comp_metrics', 'children'))
        self.outputs.append(dash.dependencies.Output('boxplo', 'figure'))
        self.outputs.append(dash.dependencies.Output('waveletplots', 'figure'))

    def execute_callbacks(self, myargs):
        self.start_time = time.time()
        self.xy_vals = {}
        op = []
        metric = myargs[-1]
        tmmu = TmMetricsUtilsUsess()
        for idx in range(len(self.pkt_pdus_dfs)):
            filter_str, conv, endpt, app, bsize = myargs[(idx*5)+1 : (idx*5)+6 ]
            #logger.debug(idx, ' : ', filter_str, conv, endpt, app)
            logger.debug('before filter pkts ' + str(idx) + ' x, y --- time elapsed = ' + str(time.time() - self.start_time))
            
            assert (self.analysis_type in ['Packets', 'Pdus'])
            if self.analysis_type == 'Packets':
                filtered_pkts = tmmu.get_filtered_df(self.pkt_dfs[idx], filter_str, conv, endpt, app)
                #logger.debug('before getting pdus ' + str(idx) + ' x, y --- time elapsed = ' + str(time.time() - self.start_time))
                #filtered_pdus = Pcap2csvutils().pdus_from_pkts(filtered_pkts)
                logger.debug('filtered pkts len %d', len(filtered_pkts))

                logger.debug('metric is %s', metric)

                x, y = tmmu.get_xy_vals(filtered_pkts, metric)
                op.append(self.get_summary(idx, filtered_pkts))#, filtered_pdus))
            else:
                filtered_pdus = tmmu.get_filtered_df(self.pdu_dfs[idx], filter_str, conv, endpt, app)
                filtered_pdus = filtered_pdus[filtered_pdus.d_size >= 0]
                #logger.debug 'filtered pdus len', len(filtered_pdus)

                logger.debug('metric is %s', metric)

                x, y = tmmu.get_xy_vals(filtered_pdus, metric)
                op.append(self.get_summary(idx, filtered_pdus))

            #print(len(x), len(y))
            logger.debug('after getting' + str(idx) + ' x, y --- time elapsed = ' + str(time.time() - self.start_time))
            self.xy_vals[idx] = [x,y]
            op.append(self.get_timeseries(x, y, metric))
            logger.debug('after getting' + str(idx) + ' time series--- time elapsed = ' + str(time.time() - self.start_time))
            
            op.append(self.get_stats(x, y, metric))
            
            logger.debug('input binsize = %s', str(bsize))
            bsize = None if (bsize == None or bsize == '') else float(bsize)
            op.extend(self.get_hist(x, y, metric, binsize = bsize)) #get_hist must return [fig, binsize]
            logger.debug('after getting' + str(idx) + ' hist --- time elapsed = ' + str(time.time() - self.start_time))
        op.append(self.get_cdfs(metric))
        logger.debug('after getting  cdf--- time elapsed = ' + str(time.time() - self.start_time))
        op.append(self.get_qqplots(metric))
        op.append(self.get_comp_metrics(metric))
        op.append(self.get_boxplots(metric))
        op.append(self.get_wavlet_scaling_plots(metric))
        logger.debug('on return x, y --- time elapsed = ' + str(time.time() - self.start_time))
        return op

    def makemydash(self):
        self.make_dashboard()
        done = False
        for i in range(8050, 8060):
            try:
                self.app.run_server(port=i, host='0.0.0.0')
                done = True
                break
            except:
                pass
        if not done:
            self.app.run_server(port=i+1, host='0.0.0.0')


class PDUDashBoard(DashBoard):
    def __init__(self, pdu_dfs, pdus_names=None):
        if pdus_names:
            assert (len(pdu_dfs) == len(pdus_names))

        self.pdu_dfs = pdu_dfs
        self.pdus_names = pdus_names
        
        for df in self.pdu_dfs:
            df.sort_values(by=['rel_time'], inplace=True) 

        self.pkt_pdus_dfs = self.pdu_dfs
        self.analysis_type = 'Pdus'
        self.gmetrics = pdu_graph_metrics
        self.implemented = pdu_graph_metrics
        for pdu_df in self.pdu_dfs:
            pdu_df['time']=pdu_df.rel_time
            pdu_df['pdu_idx']=pdu_df.idx
            pdu_df.sort_values(by=['time'], inplace=True)
            if 'l4_d_size' in pdu_df:
                pdu_df['d_size']=pdu_df.l4_d_size
        DashBoard.__init__(self,pdu_dfs)

    def get_conversations(self, pdu_df):
        pdu_df['src_ip_port'] = pdu_df['l3_src_ip'] + ':' + pdu_df['l4_src_port'].map(str)
        pdu_df['dst_ip_port'] = pdu_df['l3_dst_ip'] + ':' + pdu_df['l4_dst_port'].map(str)
        l4mask1 = pdu_df.l4_src_port < pdu_df.l4_dst_port
        pdu_df['l4_conv'] = pdu_df.loc[l4mask1, 'src_ip_port'] + ' <-> ' + pdu_df['dst_ip_port']
        pdu_df['l4_conv'].fillna(pdu_df['dst_ip_port'] + ' <-> ' + pdu_df['src_ip_port'], inplace=True)
        l4_conv_list = pdu_df.l4_conv.dropna().unique().tolist()

        ret = [{'label':'all', 'value':'all'}]
        ret.extend([{'label':'layer4: ' + x, 'value':'layer4: ' + x} for x in l4_conv_list])
        return ret

    def get_endpoints(self, pdu_df):
        l4_endpt_list = pd.unique(pdu_df[['src_ip_port', 'dst_ip_port']].dropna().values.ravel('K'))

        ret = [{'label':'all', 'value':'all'}]
        ret.extend(sum([[{'label':'layer4-rx: ' + x, 'value':'layer4-rx: ' + x}, {'label':'layer4-tx: ' + x, 'value':'layer4-tx: ' + x}]  for x in l4_endpt_list], []))
        return ret

    def get_applications(self, pdu_df):
        apps = [{'label':'all', 'value':'all'}]
        apps.extend([{'label':x, 'value':x} for x in pdu_df.app_idx.unique().tolist()])
        return apps

    def get_summary(self, idx, pdus_df):
        logger.debug('obtaining filtered pkt summary')
        
        ret = TmMetricsUtilsUsess().get_summary_pdus(pdus_df)
        r = 'PCAP ' + str(idx) + ' SUMMARY: '#]
        for name, val  in ret.items():
            #r.append(html.Br())
            r = r + '  ||  ' + name + ': ' + str(val)#)
        return r

    def make_dashboard(self):
        self.app.layout = self.get_layout() 
        self.append_dependencies()
        @self.app.callback(
            self.outputs,
            [dash.dependencies.Input('go-button', 'n_clicks')],
            state=self.input_state)
        def on_click_go(*args):
            return self.execute_callbacks(args)


class PKTDashBoard(DashBoard):
    def __init__(self, pkt_dfs, pdus_names=None):
        if pdus_names:
            assert (len(pdus_dfs) == len(pdus_names))

        self.pkt_dfs = pkt_dfs
        self.pdus_names = pdus_names

        for df in self.pkt_dfs:
            df.sort_values(by=['rel_time'], inplace=True) 

        self.pkt_pdus_dfs = self.pkt_dfs
        self.analysis_type = 'Packets'
        self.gmetrics = pkt_graph_metrics

        self.implemented = pkt_graph_metrics[:-1]
        for pkt_df in self.pkt_dfs:
            pkt_df['time']=pkt_df.rel_time
            pkt_df['pkt_idx']=pkt_df.idx
            pkt_df['d_size']=pkt_df.l4_d_size

        DashBoard.__init__(self,pkt_dfs)

    def get_conversations(self, pkt_df):
        l2mask1 = pkt_df.l2_src_mac < pkt_df.l2_dst_mac
        pkt_df['l2_conv'] = pkt_df.loc[l2mask1, 'l2_src_mac'] + ' <-> ' + pkt_df['l2_dst_mac']
        pkt_df['l2_conv'].fillna(pkt_df['l2_dst_mac'] + ' <-> ' + pkt_df['l2_src_mac'], inplace=True)
        l2_conv_list = pkt_df.l2_conv.dropna().unique().tolist()

        l3mask1 = pkt_df.l3_src_ip < pkt_df.l3_dst_ip
        pkt_df['l3_conv'] = pkt_df.loc[l3mask1, 'l3_src_ip'] + ' <-> ' + pkt_df['l3_dst_ip']
        pkt_df['l3_conv'].fillna(pkt_df['l3_dst_ip'] + ' <-> ' + pkt_df['l3_src_ip'], inplace=True)
        l3_conv_list = pkt_df.l3_conv.dropna().unique().tolist()


        pkt_df['src_ip_port'] = pkt_df['l3_src_ip'] + ':' + pkt_df['l4_src_port'].map(str)
        pkt_df['dst_ip_port'] = pkt_df['l3_dst_ip'] + ':' + pkt_df['l4_dst_port'].map(str)
        l4mask1 = pkt_df.l4_src_port < pkt_df.l4_dst_port
        pkt_df['l4_conv'] = pkt_df.loc[l4mask1, 'src_ip_port'] + ' <-> ' + pkt_df['dst_ip_port']
        pkt_df['l4_conv'].fillna(pkt_df['dst_ip_port'] + ' <-> ' + pkt_df['src_ip_port'], inplace=True)
        l4_conv_list = pkt_df.l4_conv.dropna().unique().tolist()

        trunc = {'label':'too many convs ... truncated', 'value':'too many convs ... truncated'}
        ret = [{'label':'all', 'value':'all'}]
        ret.extend([{'label':'layer2: ' + x, 'value':'layer2: ' + x} for x in l2_conv_list])
        ret.extend([{'label':'layer3: ' + x, 'value':'layer3: ' + x} for x in l3_conv_list])
        ret.extend([{'label':'layer4: ' + x, 'value':'layer4: ' + x} for x in l4_conv_list])
        #print (ret[:100] , '<<<<<<<<<<<<<')
        ret = ret[:100] + [trunc] if len(ret) > 100 else ret
        #print (ret , '<<<<<<<<<<<<<')
        return ret

    def get_endpoints(self, pkt_df):
        l2_endpt_list = pd.unique(pkt_df[['l2_src_mac', 'l2_dst_mac']].dropna().values.ravel('K'))
        l3_endpt_list = pd.unique(pkt_df[['l3_src_ip', 'l3_dst_ip']].dropna().values.ravel('K'))
        l4_endpt_list = pd.unique(pkt_df[['src_ip_port', 'dst_ip_port']].dropna().values.ravel('K'))

        trunc = {'label':'too many endpts ... truncated', 'value':'too many endpts ... truncated'}
        ret = [{'label':'all', 'value':'all'}]
        ret.extend(sum([[{'label':'layer2-rx: ' + x, 'value':'layer2-rx: ' + x}, {'label':'layer2-tx: ' + x, 'value':'layer2-tx: ' + x}]  for x in l2_endpt_list], []))
        ret.extend(sum([[{'label':'layer3-rx: ' + x, 'value':'layer3-rx: ' + x}, {'label':'layer3-tx: ' + x, 'value':'layer3-tx: ' + x}]  for x in l3_endpt_list], []))
        ret.extend(sum([[{'label':'layer4-rx: ' + x, 'value':'layer4-rx: ' + x}, {'label':'layer4-tx: ' + x, 'value':'layer4-tx: ' + x}]  for x in l4_endpt_list], []))
        ret = ret[:100] + [trunc]  if len(ret) > 100 else ret
        return ret

    def get_applications(self, pkt_df):
        apps = [{'label':'all', 'value':'all'}]
        trunc = {'label':'too many apps ... truncated', 'value':'too many apps ... truncated'}
        apps.extend([{'label':x, 'value':x} for x in pkt_df.app_idx.unique().tolist()])
        apps = apps[:100] + [trunc] if len(apps)>100 else apps
        return apps

    def get_summary(self, idx, pkts_df):#, pdus_df):
        logger.debug('obtaining filtered pkt summary')
        
        ret = TmMetricsUtilsUsess().get_summary_pkts(pkts_df)#, pdus_df)
        r = 'PCAP ' + str(idx) + ' SUMMARY: '#]
        for name, val  in ret.items():
            #r.append(html.Br())
            r = r + '  ||  ' + name + ': ' + str(val)#)
        return r

    def make_dashboard(self):
        self.app.layout = self.get_layout() 
        self.append_dependencies()
        @self.app.callback(
            self.outputs,
            [dash.dependencies.Input('go-button', 'n_clicks')],
            state=self.input_state)
        def on_click_go(*args):
            return self.execute_callbacks(args)
