import os
import sys
import json
import math
import logging
import textwrap
import pandas as pd
import statistics as st
import numpy as np
import matplotlib.pyplot as plt
from collections import OrderedDict
from .const import conn_pkts_reqd_metrics, pkt_graph_metrics, pdu_graph_metrics
from .const import per_pkt_metrics, metric_units, time_resolution

#from trafficmodelingutils.trafficmodelingutils import Trafficmodelingutils

import matplotlib
matplotlib.rcParams.update({'font.size': 12})

logging.getLogger("matplotlib").setLevel(logging.INFO)

mark = ['bo-', 'g^-', 'rx-', 'kd-', 'm+-']

tempstart = '''
            \\title{Traffic Modeling Report}
            \\author{UH Netlab}
            \\date{\\today}
            \\documentclass[10pt]{article}

            \\usepackage{graphicx}

            \\begin{document}
            \\maketitle
            \\hspace{0pt}
            \\vfill
            \\section{Introduction}
            In each of the sections below we provide the charts for each app in the input traces 
            \\vfill
            \\hspace{0pt}
            \\newpage
            '''

tempend = '''
            \\hspace{0pt}
            \\vfill
            \\section{Conclusions}\\label{conclusions}
            Thank you
            \\vfill
            \\hspace{0pt}
            \\end{document}
            '''

def fmt(val):
    if isinstance(val,float):
        return float('%.3f' % (val))
    return val

def removeOutliers(x, outlierConstant=7):
    if len(x) == 0:
        return x
    a = np.array(x)
    upper_quartile = np.percentile(a, 75)
    lower_quartile = np.percentile(a, 25)
    #print 'upper_quartile', upper_quartile
    #print 'lower_quartile', lower_quartile
    IQR = (upper_quartile - lower_quartile) * outlierConstant
    quartileSet = (lower_quartile - IQR, upper_quartile + IQR)
    xpd = pd.Series(a)
    remlist = xpd[xpd.between(quartileSet[0], quartileSet[1])].tolist()
    outlist = xpd[~xpd.between(quartileSet[0], quartileSet[1])].tolist()
    #print 'remlist', len(remlist)
    #print 'outlist', len(outlist)
    return remlist, outlist

def getOutlierLimits(x, outlierConstant=7):
    if len(x) == 0:
        return None, None
    a = np.array(x)
    upper_quartile = np.percentile(a, 75)
    lower_quartile = np.percentile(a, 25)
    #print 'upper_quartile', upper_quartile
    #print 'lower_quartile', lower_quartile
    IQRR = (upper_quartile - lower_quartile) * outlierConstant
    return lower_quartile - IQRR, upper_quartile + IQRR


class Datavalidation:

    def __init__(self):
        pass

    def make_line_plot(self, usess_xys, bins, pathp, app, metric, pkts_pdus):
        #Lines

        fig, axs = plt.subplots(len(pkts_pdus), 1, sharey=True, sharex=True, constrained_layout=True, figsize=(12, 8))
        #fig.suptitle(app + ' ' + metric + ' | self-validation')
        fig.suptitle(metric)
        max1x = 0
        for i, name in enumerate(pkts_pdus):
            curr_axs = axs[i] if len(pkts_pdus) > 1 else axs # required bcos of mathplot lib sublots(1,1) implementation
            if name in usess_xys:
                for j, usess in enumerate(usess_xys[name]):
                    xx,yy = usess
                    if len(xx) > 0:
                        xx = (np.array(xx)-xx[0]).tolist()
                        #print(name, datasets[name])
                        curr_axs.bar(xx,yy, label='usess'+str(j))
                    
                    if i == 0:
                        #print(xx[-1])
                        max1x = max(max1x, xx[-1])
                #print('max1x', max1x)
                curr_axs.set_xlim(0, max1x)
                n = name.replace('modeltype', 'uhgeneric_v')
                curr_axs.set_title(n)
                curr_axs.set_xlabel('time, (seconds)')
                curr_axs.set_ylabel(metric+ ' \n(' + metric_units[metric]+')')
                #curr_axs.tick_params(axis='x', labelrotation=45)
                #curr_axs.legend()
        loc = pathp + '/' + metric + '_1line.pdf'
        plt.savefig(loc)
        #plt.show()
        plt.close(fig)
        return loc

    def make_histograms(self, dataset_list, bins, pathp, app, metric, pkts_pdus):
        #Histograms

        suptitle = metric.replace('_', ' ').capitalize() + ": Histograms for \n" + ' & '.join(pkts_pdus)
        #suptitle = metric
        #suptitle = app + ' ' + metric + ' | actual_data'

        w = len(pkts_pdus)*4
        fig, axs = plt.subplots(1, len(pkts_pdus), sharey=True, sharex=True, constrained_layout=True, figsize=(w, 4.5))
        sut = fig.suptitle(suptitle, y=0.01)
        for i, name in enumerate(pkts_pdus):
            curr_axs = axs[i] if len(pkts_pdus) > 1 else axs# required bcos of mathplot lib sublots(1,1) implementation
            n = name.replace('modeltype', 'uhgeneric_v')
            n = n.replace('-', ' ')
            #curr_axs.set_title(n)
            curr_axs.set_xlabel(metric+ ' \n(' + metric_units[metric]+')')
            curr_axs.set_ylabel('Frequency')

            if len(dataset_list[i]) >= 2:
                #print(i, '****', dat[i], '****', bins)
                curr_axs.hist(dataset_list[i], bins=bins)
                l,r = curr_axs.get_xlim()
                if l < 0:
                    curr_axs.set_xlim(0, r)
                
                sd = str(fmt(st.stdev(dataset_list[i]))) if len(dataset_list[i]) > 1 else 0
                text = n + '\n-------------- \nmean:'+str(fmt(st.mean(dataset_list[i])))+' \nmedian:'+str(fmt(st.median(dataset_list[i])))+' \nSD:'+str(fmt(sd))
                wraptitle = textwrap.TextWrapper(width=15,
                                replace_whitespace=False).fill(text)
                curr_axs.text(0.48, 0.95, wraptitle, fontsize=10, ha='left', va='top', transform=curr_axs.transAxes)
            curr_axs.tick_params(axis='x', labelrotation=45)
        loc = pathp + '/' + metric + '_2hist.pdf'
        plt.savefig(loc, 
            #bbox_extra_artists=(sut), 
            bbox_inches='tight')
        #plt.show()
        plt.close(fig)
        return loc

    def make_box_cdfs(self, dat, bins, pathp, app, metric, pkts_pdus):
        #Boxplots & CDF
        suptitle = metric.replace('_', ' ').capitalize() + ": Box plots and CDFs for \n" + '&'.join(pkts_pdus)
        #suptitle = metric

        box_w = max(4,len(pkts_pdus)*2.2)
        cdf_w = 7 if len(dat) > 2 else 5
        w = box_w + cdf_w
        fig, axu = plt.subplots(1, 2, constrained_layout=True, figsize=(w, 4.5), gridspec_kw={'width_ratios': [box_w, cdf_w]})
        
        #fig.suptitle(app + ' ' + metric + ' | actual_data')
        sut = fig.suptitle(suptitle, y=-0.01)
        labs = [n.replace('-','\n') for n in pkts_pdus]
        labs = [m.replace('modeltype', 'uhgeneric_v') for m in labs]
        axu[0].boxplot(dat, labels=labs)
        axu[0].set_ylabel(metric+ ' \n(' + metric_units[metric]+')')
        i = 0
        for i, name in enumerate(pkts_pdus):
            #print (i, name)
            n = name.replace('modeltype', 'uhgeneric_v')
            if len(dat[i]) > 0:
                #yy, rem = removeOutliers(dat[i])
                yy = dat[i]
                xx = np.array(range(0,101, 5), dtype=float)
                yy = np.percentile(yy, xx)
                axu[1].plot(yy, xx/100,  mark[i], label=n,)
                i += 1
        #print()
        axu[1].tick_params(axis='x', labelrotation=45)
        xax = axu[1].set_xlabel(metric+ ' \n(' + metric_units[metric]+')')
        axu[1].set_ylabel('F(x)')
        lg = axu[1].legend(fontsize=10,  loc='best', bbox_to_anchor=(0.5, 0., 0.5, 0.5))
        if len(pkts_pdus) > 2:
            lg = axu[1].legend(fontsize=10,   loc='center left', bbox_to_anchor=(1,0.5))
        loc = pathp + '/' + metric + '_3box_cdf.pdf'
        plt.savefig(loc, 
            bbox_extra_artists=(lg,sut), 
            bbox_inches='tight')
        #plt.show()
        plt.close(fig)
        return loc

    def get_tab(self, datasets, bins, pathp, app, metric, pkts_pdus):
        ttx =   '''
                    \\begin{table}[h!]
                    \\centering
                    \\begin{tabular}{p{3.5cm}|p{1.5cm}p{1.5cm}p{1.5cm}p{1.5cm}}
                        \\hline
                             & mean &  median  & SD & Datalen\\\\
                        \\hline
                '''
        
        for i, name in enumerate(pkts_pdus):
            if name in datasets:
                dati = datasets[name]
                me, med, sd, dlen = ['','','','']
                if len(datasets[name]) > 0:
                    me = str(fmt(st.mean(dati)))
                    med = str(fmt(st.median(dati)))
                    sd = str(fmt(st.stdev(dati))) if len(dati) > 1 else 0
                    dlen = str(len(dati))
                ttx += '''
                            %s & %s & %s & %s & %s \\\\
                        ''' %(name, me, med, sd, dlen)

        ttx +=  '''
                    \\hline 
                \\end{tabular}
                \\end{table}
                '''

        return ttx

class Selfvalidation:

    def __init__(self):
        pass

    def make_line_plot(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        #Lines
        fig, axs = plt.subplots(1, 1, sharey=True, sharex=True, constrained_layout=True, figsize=(20, 4))
        #fig.suptitle(app + ' ' + metric + ' | ' + stat + ' | self-validation')
        fig.suptitle(metric + ': pairwise user-sessions KS-2-Sample s-statistics')
        for i, name in enumerate(pkts_pdus):
            
            if name in datasets:
                if len(datasets[name]) > 0:
                    #print(name, datasets[name])

                    n = name.replace('modeltype', 'uhgeneric_v')
                    axs.plot(datasets[name],  mark[i][:1], label=n)
                #axs[i].set_title(name)
                axs.set_xlabel('Pair User-sessions #')
                axs.set_ylabel(stat)
                #axs[i].tick_params(axis='x', labelrotation=45)
        axs.legend()
        loc = pathp + '/' + metric + '_1line.pdf'
        plt.savefig(loc)
        #plt.show()
        plt.close(fig)
        return loc

    def make_histograms(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        #Histograms
        
        suptitle = metric.replace('_', ' ').capitalize() + ": Histograms of \npair-wise user-sessions KS-2-sample s-statistics"
        w = len(pkts_pdus)*4
        fig, axs = plt.subplots(1, len(pkts_pdus), sharey=True, sharex=True, constrained_layout=True, figsize=(w, 4.5))
        #fig.suptitle(app + ' ' + metric + ' | ' + stat + ' | self-validation')
        sut = fig.suptitle(suptitle, y=0.01)
        for i, name in enumerate(pkts_pdus): 
            n = name.replace('modeltype', 'uhgeneric_v')
            n = n.replace('-', ' ')
            curr_axs = axs[i] if len(pkts_pdus) > 1 else axs  # required bcos of mathplot lib sublots(1,1) implementation
            if name in datasets:
                dat = datasets[name]
                #print(name, dat)
                
                if 'conn_' in metric or '_throughput' in metric:
                    print('????? hist ',name,  len(dat))
                #curr_axs.set_title(n)
                curr_axs.set_xlabel('KS-2-sample S statistic')
                curr_axs.set_ylabel('Frequency')
                if len(dat)>0:
                    curr_axs.hist(dat, bins=bins)
                    sd = str(fmt(st.stdev(dat))) if len(dat) > 1 else 0
                    text = n + '\n-------------- \nmean:'+str(fmt(st.mean(dat)))+' \nmedian:'+str(fmt(st.median(dat)))+' \nSD:'+str(fmt(sd))
                    wraptitle = textwrap.TextWrapper(width=15,
                                replace_whitespace=False).fill(text)
                    curr_axs.text(0.48, 0.95, wraptitle, fontsize=10, ha='left', va='top', transform=curr_axs.transAxes)
                curr_axs.tick_params(axis='x', labelrotation=45)
        loc = pathp + '/' + metric + '_2hist.pdf'
        plt.savefig(loc, 
            #bbox_extra_artists=(sut), 
            bbox_inches='tight')
        #plt.show()
        plt.close(fig)
        return loc

    def make_box_cdfs(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        #Boxplots & CDF
        suptitle = metric.replace('_', ' ').capitalize() + ": Box plots and CDFs of \npair-wise user-sessions KS-2-sample s-statistics"

        # if 'conn_arrival_rate' in metric:
        #     print(metric)
        #     print(datasets.keys())
        #     print(datasets['uhgeneric_v4-traffic'])

        box_w = max(4,len(pkts_pdus)*2.2)
        cdf_w = 7 if len(datasets) > 2 else 5
        w = box_w + cdf_w
        fig, axu = plt.subplots(1, 2, constrained_layout=True, figsize=(w, 4.5), gridspec_kw={'width_ratios': [box_w, cdf_w]})
        
        #fig.suptitle(app + ' ' + metric + ' | ' + stat + ' | self-validation')
        sut = fig.suptitle(suptitle, y=-0.01)
        labs = [n.replace('-','\n') for n in pkts_pdus]
        labs = [m.replace('modeltype', 'uhgeneric_v') for m in labs]
        #print (len(datasets.values()), labs)
        axu[0].boxplot(list(datasets.values()), labels=labs)
        axu[0].set_ylabel('KS-2-sample S statistic')
        i = 0
        #for name, data in datasets.items():
        for i, name in enumerate(pkts_pdus):
            n = name.replace('modeltype', 'uhgeneric_v')
            if name not in datasets:
                continue
            data = datasets[name]
            if len(data) > 0:
                x = np.array(range(0,101, 5))
                y = np.percentile(data, x)

                if 'conn_' in metric or '_throughput' in metric:
                    print('????? box ',name,  len(data))
                #print('y', y)
                axu[1].plot(y, x/100.0, mark[i], label=n)
                i += 1
        lg = axu[1].legend(fontsize=10,  loc='best', bbox_to_anchor=(0.5, 0., 0.5, 0.5))
        if len(datasets) > 2:
            lg = axu[1].legend(fontsize=10,   loc='center left', bbox_to_anchor=(1,0.5))
        xax = axu[1].set_xlabel('KS-2-sample S statistic')
        axu[1].set_ylabel('F(x)')
        loc = pathp + '/' + metric + '_3box_cdf.pdf'
        plt.savefig(loc, 
            bbox_extra_artists=(lg,sut), 
            bbox_inches='tight'
            )
        #plt.show()
        plt.close(fig)
        return loc

    def get_tab(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        ttx =   '''
                    \\begin{table}[h!]
                    \\centering
                    \\begin{tabular}{p{3.5cm}|p{1.5cm}p{1.5cm}p{1.5cm}p{1.5cm}}
                        \\hline
                             & mean &  median  & SD & Datalen\\\\
                        \\hline
                '''

        
        for i, name in enumerate(pkts_pdus):
            if name in datasets:
                dat = datasets[name]
                me, med, sd, dlen = ['','','','']
                if len(datasets[name]) > 0:
                    me = str(fmt(st.mean(dat)))
                    med = str(fmt(st.median(dat)))
                    sd = str(fmt(st.stdev(dat))) if len(dat) > 1 else 0
                    dlen = str(len(dat))
                ttx += '''
                            %s & %s & %s & %s & %s \\\\
                        ''' %(name, me, med, sd, dlen)

        ttx +=  '''
                    \\hline 
                \\end{tabular}
                \\end{table}
                '''

        return ttx


class Crossvalidation:

    def __init__(self):
        pass

    def make_line_plot(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        #Lines
        fig, axs = plt.subplots(1, 1, sharey=True, sharex=True, constrained_layout=True, figsize=(12, 5))
        fig.suptitle(app + ' ' + metric + ' | ' + stat + ' | self-validation')
        for i, name in enumerate(pkts_pdus):
            if name in datasets:
                if len(datasets[name]) > 0:
                    #print(name, datasets[name])
                    axs.plot(datasets[name],  mark[i][:1], label=name)
                #axs[i].set_title(name)
                axs.set_xlabel('run comparison #')
                axs.set_ylabel(stat)
                #axs[i].tick_params(axis='x', labelrotation=45)
        axs.legend()
        loc = pathp + '/' + metric + '_1line.pdf'
        plt.savefig(loc, 
            bbox_extra_artists=(lg,sut), 
            bbox_inches='tight')
        #plt.show()
        plt.close(fig)
        return loc

    def make_histograms(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        #Histograms
        suptitle = metric.replace('_', ' ').capitalize() + ": Histograms of \ncross validation by user-sessions " + stat
        
        fig, axs = plt.subplots(1, len(pkts_pdus), sharey=True, sharex=True, constrained_layout=True, figsize=(13, 5.5))
        sut = fig.suptitle(suptitle, y=-0.01)
        for i, name in enumerate(pkts_pdus):
            curr_axs = axs[i] if len(pkts_pdus) > 1 else axs
            if name in datasets:
                dat = datasets[name]
                n = name.replace('modeltype', 'uhgeneric_v')
                n = n.replace('-', ' ')
                #print(name, dat)
                #curr_axs.set_title(name)
                curr_axs.set_xlabel(stat)
                curr_axs.set_ylabel('Frequency')
                if len(dat)>0:
                    curr_axs.hist(dat, bins=bins)
                    mysd = st.stdev(dat) if len(dat)>1 else 0
                    text = n + '\n-------------- \nmean:'+str(fmt(st.mean(dat)))+' \nmedian:'+str(fmt(st.median(dat)))+' \nSD:'+str(fmt(mysd))
                    wraptitle = textwrap.TextWrapper(width=15,
                                replace_whitespace=False).fill(text)
                    curr_axs.text(0.48, 0.95, wraptitle, fontsize=10, ha='left', va='top', transform=curr_axs.transAxes)
                curr_axs.tick_params(axis='x', labelrotation=45)
        loc = pathp + '/' + metric + '_2hist.pdf'
        plt.savefig(loc, 
            #bbox_extra_artists=(sut), 
            bbox_inches='tight')
        #plt.show()
        plt.close(fig)
        return loc

    def make_box_cdfs(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        #Boxplots & CDF
        suptitle = metric.replace('_', ' ').capitalize() + ": Boxplots of cross validation by  \nuser-sessions " + stat
         
        box_w = max(4,len(pkts_pdus)*2.2)
        cdf_w = 7 if len(datasets) > 2 else 5
        w = box_w + cdf_w
        fig, axu = plt.subplots(1, 2, constrained_layout=True, figsize=(w, 4.5), gridspec_kw={'width_ratios': [box_w, cdf_w]})
        
        sut = fig.suptitle(suptitle, y=-0.01)
        labs = [n.replace('-','\n') for n in pkts_pdus]
        #print (len(datasets.values()), labs)
        axu[0].boxplot(list(datasets.values()), labels=labs)
        axu[0].set_ylabel(stat)
        i = 0
        for name, data in datasets.items():
            if len(data) > 0:
                x = np.array(range(0,101, 5))
                y = np.percentile(data, x)

                #print('y', y)
                axu[1].plot(y, x/100.0, mark[i], label=name)
                i += 1
        lg = axu[1].legend(fontsize=14,  loc='best', bbox_to_anchor=(0.5, 0., 0.5, 0.5))
        if len(datasets) > 2:
            lg = axu[1].legend(fontsize=14,   loc='upper center', bbox_to_anchor=(0.5,-0.4))
        xax = axu[1].set_xlabel('stat')
        axu[1].set_ylabel('F(x)')
        loc = pathp + '/' + metric + '_3box_cdf.pdf'
        plt.savefig(loc, 
            bbox_extra_artists=(lg,sut), 
            bbox_inches='tight')
        #plt.show()
        plt.close(fig)
        return loc

    def get_tab(self, datasets, bins, pathp, app, metric, stat, pkts_pdus):
        ttx =   '''
                    \\begin{table}[h!]
                    \\centering
                    \\begin{tabular}{p{3.5cm}|p{1.5cm}p{1.5cm}p{1.5cm}p{1.5cm}}
                        \\hline
                             & mean &  median  & SD & Datalen\\\\
                        \\hline
                '''

        
        for i, name in enumerate(pkts_pdus):
            if name in datasets:
                dat = datasets[name]
                me, med, sd, dlen = ['','','','']
                if len(datasets[name]) > 0:
                    me = str(fmt(st.mean(dat)))
                    med = str(fmt(st.median(dat)))
                    sd = str(fmt(st.stdev(dat)))
                    dlen = str(len(dat))
                ttx += '''
                            %s & %s & %s & %s & %s \\\\
                        ''' %(name, me, med, sd, dlen)

        ttx +=  '''
                    \\hline 
                \\end{tabular}
                \\end{table}
                '''

        return ttx
