# Copyright (c) 2017  University of Houston

from __future__ import print_function

import os
import sys
import math
import time
import json
import math
import logging
import argparse
import shutil
import warnings
from collections import OrderedDict
import numpy as np
import pandas as pd
import statistics as st
from pcap2csv.pcap2csv import Pcap2csv
from pcap2csv.pcap2csvutils import Pcap2csvutils
from .dashboard import DashBoard
from .const import conn_pkts_reqd_metrics, pkt_graph_metrics, pdu_graph_metrics, conn_graph_metrics
from .const import per_pkt_metrics, metric_units, time_resolution
from .tmmetricsutilsusess import TmMetricsUtilsUsess
from .plotter import Selfvalidation, Crossvalidation, Datavalidation

tempstart = '''
            \\title{Traffic Modeling Report\\report_type}
            \\author{UH Netlab}
            \\date{\\today}
            \\documentclass[10pt]{article}

            \\usepackage{graphicx}

            \\begin{document}
            \\maketitle
            \\hspace{0pt}
            \\vfill
            \\section{Introduction}
            In each of the sections below we provide the charts for each app in the input traces 
            \\vfill
            \\hspace{0pt}
            \\clearpage
            '''

tempend = '''
            \\hspace{0pt}
            \\vfill
            \\section{Conclusions}\\label{conclusions}
            Thank you
            \\vfill
            \\hspace{0pt}
            \\end{document}
            '''

try:
    warnings.simplefilter(action="ignore", category=pd.core.common.SettingWithCopyWarning)
except AttributeError:
    warnings.simplefilter(action="ignore", category=pd.errors.SettingWithCopyWarning)

logger = logging.getLogger(__name__)

def fmt(val):
    if isinstance(val,float):
        return float('%.4f' % (val))
    return val

def round_to_1(x):
    if x==0:
        return x
    return round(x, -int(math.floor(math.log10(abs(x)))))

def update_rounded(d1, d2):
    for k, v in d2.items():
        if isinstance(v, float):
            d1[k] = float('%.5f' % (v))
        else:
            d1[k] = v
    return d1

def myconverter(obj):
    if isinstance(obj, dict) or isinstance(obj, OrderedDict):
        for key,val in obj.items():
            obj[key]=myconverter(val)
        return obj
    elif isinstance(obj, list):
        for val in obj:
            val=myconverter(val)
        return obj
    elif isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    else:
        return obj

def getOutlierLimits(x, outlierConstant=7):
    if len(x) == 0:
        return None, None
    a = np.array(x)
    upper_quartile = myconverter(np.percentile(a, 75))
    lower_quartile = myconverter(np.percentile(a, 25))
    #print 'upper_quartile', upper_quartile
    #print 'lower_quartile', lower_quartile
    IQRR = (upper_quartile - lower_quartile) * outlierConstant
    return myconverter(lower_quartile - IQRR), myconverter(upper_quartile + IQRR)

def get_bins(min_y, max_y, nbins=20):
    #binning
    min_y=0 if not min_y else min_y
    max_y=0 if not max_y else max_y

    bin_width = (max_y - min_y)/nbins
    bin_width = round_to_1(bin_width)
    miniy = min_y//bin_width*bin_width if bin_width !=0 else 0
    maxiy = (max_y//bin_width*bin_width) + bin_width + bin_width if bin_width !=0 else 0
    bins = np.arange(miniy, maxiy , bin_width) if bin_width !=0 else []
    #print('mini:', miniy, ',  maxi:', maxiy, ',  bin_width:', bin_width, ',   bins:', bins)

    if len(bins) == 0:
        bins = [i/float(nbins) for i in range(nbins)]
    return bins

class TrafficMetricsTextUsess():
    '''analyzes network traffic files to present a JSON report of metrics
    and perform comparison'''
    def __init__(self, dfs, outdir, analysis_type, analysis_kind, dfs_types,
                 dfs_names=None,  ):
        self.dfs = dfs
        self.outdir = outdir
        self.analysis_type = analysis_type
        self.analysis_kind = analysis_kind
        self.dfs_types = dfs_types
        self.dfs_names = dfs_names

        self.start_time = 0
        self.gmetrics = None
        self.pdus_dfs = []
        self.pkts_dfs = []

        #print(dfs_types)
        assert (analysis_type in ['packets', 'pdus', 'both'])
        assert (analysis_kind in ['global', 'per_user_session', 'both'])
        for dft in dfs_types:
            assert (dft in ['packets-csv', 'pdus-csv'])
        if dfs_names:
            assert (len(dfs) == len(dfs_names))
        else:
            self.dfs_names = \
                [self.analysis_type + '-' + str(i) for i in range(len(dfs))]
        for df in self.dfs:
            df.sort_values(by=['rel_time'], inplace=True) 
            if 'd_size' not in df:
                df['l4_d_size'] = pd.to_numeric(df['l4_d_size'],
                                                errors='coerce')
                df['d_size'] = df['l4_d_size']
            df['d_size'] = pd.to_numeric(df['d_size'], errors='coerce')
            df['time']=df.rel_time
        
    def none_in_dflist(self, dflist):
        for i in dflist:
            if i is None:
                return True
        return False

    def get_gmetrics(self):
        if self.analysis_type == 'packets':
            if self.none_in_dflist(self.pkts_dfs):
                raise ValueError('cannot analyze packets, when input type is pdu-csv')
            self.gmetrics = pkt_graph_metrics + conn_graph_metrics
        elif self.analysis_type == 'pdus':
            self.gmetrics = pdu_graph_metrics + conn_graph_metrics
        elif self.analysis_type == 'both':
            if self.none_in_dflist(self.pkts_dfs):
                raise ValueError('cannot analyze both packets and pdus, when input type is pdu-csv')
            self.gmetrics = pkt_graph_metrics + pdu_graph_metrics + conn_graph_metrics
            logger.debug('list of metrics is %s', str(self.gmetrics) )
        else:
            raise ValueError('Invalid analysis type')

    def get_pkts_pdus(self):
        for i, df in enumerate(self.dfs):
            #print(self.dfs_types[i])
            if self.dfs_types[i] == 'packets-csv':
                self.pkts_dfs.append(df)
                self.pdus_dfs.append(Pcap2csvutils().pdus_from_pkts(df))
            else:
                self.pkts_dfs.append(None)
                self.pdus_dfs.append(df)

    def run(self):
        self.get_pkts_pdus()
        self.get_gmetrics()

        logger.debug('analysis type is %s', self.analysis_kind)

        report, report_usess = [None, None]
        if self.analysis_kind in ['global', 'both']:
            logger.debug('running global analysis')
            report = self.run_global()
        
        if self.analysis_kind in ['per_usess', 'both']:
            logger.debug('running per user session usess analysis')
            report_usess = self.run_per_usess()

        return report, report_usess

    def run_global(self):
        tmmu_usess = TmMetricsUtilsUsess()
        report = OrderedDict()
        report['summaries'] = OrderedDict()
        report['apps'] = OrderedDict()

        for i, df in enumerate(self.dfs):
            #print '------', i, len(df.index)
            report['summaries'][self.dfs_names[i]] = OrderedDict()
            reportx = report['summaries'][self.dfs_names[i]]
            if self.analysis_type in ['pkts', 'both']:
                update_rounded(reportx, 
                               tmmu_usess.get_summary_pkts(self.pkts_dfs[i]))
            if self.analysis_type in ['pdus', 'both']:
                update_rounded(reportx,
                               tmmu_usess.get_summary_pdus(self.pdus_dfs[i]))

        #apps = pd.concat(self.dfs).app_idx.dropna().unique().tolist()
        apps = ['allapps'] + pd.concat(self.dfs).app_idx.dropna().\
               unique().tolist()
        for app in apps:
            report['apps'][app] = OrderedDict()

            for metric in self.gmetrics:
                logger.debug('processing metric ==== %s', metric)
                report['apps'][app][metric] = OrderedDict()
                xy_vals = OrderedDict()
                for i in range(len(self.dfs)):
                    if metric in pkt_graph_metrics:
                        df = self.pkts_dfs[i]
                    else:
                        df = self.pdus_dfs[i]

                    if app == 'allapps' or app == 'all':
                        app_df = df
                    else:
                        app_df = df.loc[df['app_idx'] == app]

                    if len(app_df) == 0:
                        xy_vals[i] = ([], [])
                        continue
                    report['apps'][app][metric][self.dfs_names[i]] = OrderedDict()
                    reportx = report['apps'][app][metric][self.dfs_names[i]]
                    x, y = tmmu_usess.get_xy_vals(app_df, metric)
                    stats = tmmu_usess.get_stats(x, y, metric)
                    update_rounded(reportx, stats)
                    xy_vals[i] = (x, y)

                    #if self.store_xy_vals:
                    #reportx['x_vals'] = x
                    #reportx['y_vals'] = y
                    #reportx['xy_vals'] = (x, y)

                report['apps'][app][metric]['comparison'] = OrderedDict()
                reporty = report['apps'][app][metric]['comparison']
                for j in range(len(self.dfs)-1):
                    for k in range(j+1, len(self.dfs)):
                        x1, y1 = xy_vals[j]
                        x2, y2 = xy_vals[k]
                        name = self.dfs_names[j] + '-vs-' + self.dfs_names[k]
                        reporty[name] = tmmu_usess.get_comp_metrics(y1, y2,
                                                                    metric)
        return report

    def run_per_usess(self):
        
        tex_datav = tempstart.replace('report_type', 'First Order Data Validtion')
        tex_selfv = tempstart.replace('report_type', 'Second Order Self Validtion')
        tex_crossv = tempstart.replace('report_type', 'Cross Validtion')

        tmmu_usess = TmMetricsUtilsUsess()
        report_usess = OrderedDict()
        apps = pd.concat(self.dfs).app_idx.dropna().unique().tolist()
        #apps = ['allapps'] + pd.concat(self.dfs).app_idx.dropna().\
        #        unique().tolist()
                


        report_usess['summaries'] = OrderedDict()
        for i, df in enumerate(self.dfs):
            #print '------', i, len(df.index)
            report_usess['summaries'][self.dfs_names[i]] = OrderedDict()
            reportx = report_usess['summaries'][self.dfs_names[i]]
            if self.analysis_type in ['pkts', 'both']:
                update_rounded(reportx, 
                               tmmu_usess.get_summary_pkts_usess(self.pkts_dfs[i]))
            if self.analysis_type in ['pdus', 'both']:
                update_rounded(reportx,
                               tmmu_usess.get_summary_pdus_usess(self.pdus_dfs[i]))

        apps = pd.concat(self.dfs).app_idx.dropna().unique().tolist()
        #apps = ['allapps'] + pd.concat(self.dfs).app_idx.dropna().\
        #       unique().tolist()

        for app in apps:
            appi = app.replace('_', '')
            appi = appi.replace('-', '')
            atex =  '''\n \\hspace{0pt} \n \\vfill \n \section{%s}\label{%s}
                        In this section we describe Application %s.
                        \\vfill \n \\hspace{0pt} \n \\clearpage \n'''%(appi, appi,appi)
            tex_datav += atex
            tex_selfv += atex
            tex_crossv += atex

            report_usess[app] = OrderedDict()

            for metric in self.gmetrics:
                m = metric.replace('_', ' ')
                mtex = '''\n \subsection{%s}\label {%s}\n ''' %(m, m)
                tex_datav += mtex
                tex_selfv += mtex
                tex_crossv += mtex

                logger.debug(metric)
                report_usess[app][metric] = OrderedDict()
                all_usess_xys = OrderedDict()

                #get x,y values
                for i, name in enumerate(self.dfs_names):
                    if metric in pkt_graph_metrics:
                        df = self.pkts_dfs[i]
                    else:
                        df = self.pdus_dfs[i]
                    if app == 'allapps' or app == 'all':
                        app_df = df
                    else:
                        app_df = df.loc[df['app_idx'] == app]
                    if len(app_df) == 0:
                        all_usess_xys[name] = ()
                        continue
                    usess_xys = tmmu_usess.get_xy_vals_usess(app_df, metric)
                    all_usess_xys[name] = usess_xys

                #remove outliers
                all_ys = []
                for name, usess_xys in all_usess_xys.items():
                    for usess in usess_xys:
                        x,y=usess
                        all_ys.extend(y)
                outl_thresh_low, outl_thresh_high = getOutlierLimits(all_ys)
                for name, usess_xys in all_usess_xys.items():
                    for i in range(len(usess_xys)):
                        x,y=usess_xys[i]
                        dfxy = pd.DataFrame(list(zip(x, y)), 
                                          columns =['x', 'y'])
                        dfxy = dfxy[(dfxy['y'] >= outl_thresh_low) & (dfxy['y'] <= outl_thresh_high)]
                        all_usess_xys[name][i] = (dfxy.x.tolist(), dfxy.y.tolist())

                #get stats & update usess report
                for name, usess_xys in all_usess_xys.items():
                    usess_stats_distr = tmmu_usess.get_stats_usess(usess_xys,
                                                                   metric)
                    report_usess[app][metric][name] = OrderedDict()
                    reportx_usess = report_usess[app][metric][name]
                    update_rounded(reportx_usess, usess_stats_distr)

                #compare and update usess report
                report_usess[app][metric]['comparison'] = OrderedDict()
                reporty_usess = report_usess[app][metric]['comparison']
                for j in range(len(self.dfs)):
                    namej = self.dfs_names[j] 
                    for k in range(j+1, len(self.dfs)):
                        namek = self.dfs_names[k] 
                        name = namej + '-vs-' + namek
                        reporty_usess[name] = tmmu_usess.get_comp_metrics_usess(
                            all_usess_xys[namej], all_usess_xys[namek], metric)   
                    break # remove for full pairwise compare

                tex_datav += self.plot_datavalidation(app, metric, all_usess_xys)
                tex_selfv += self.plot_selfvalidation(app, metric, report_usess[app][metric])
                if len(self.dfs_names)>1:
                    tex_crossv += self.plot_crossvalidation(app, metric, report_usess[app][metric])

        texpairs =[(tex_datav, 'data'), (tex_selfv, 'self')]
        if len(self.dfs_names)>1:
            texpairs.append((tex_crossv, 'cross'))

        for ti, name in texpairs:
            ti += tempend  
            texfile = self.outdir + '/pdf_report_' + name + 'validation.tex'
            with open(texfile, 'w') as f:
                f.write(ti)

            cur_dir = os.getcwd()
            os.chdir(self.outdir)
            os.system('pdflatex -interaction=nonstopmode pdf_report_'+name+'validation.tex ')
            os.chdir(cur_dir)

        return report_usess

    def plot_datavalidation(self, app, metric, all_usess_xys):
        mark = ['bo-', 'g^-', 'rx-', 'kd-', 'm+-']

        tex = ''

        dataset_list, all_ys = [], []
        for i, (name, usess_xys) in enumerate(all_usess_xys.items()):
            #print(name)
            dataset_list.append([])
            for usess in usess_xys:
                x,y=usess
                dataset_list[i].extend(y)
                all_ys.extend(y)

        if len(all_ys) > 0:
            min_y, max_y = min(all_ys), max(all_ys)
            bins = get_bins( min_y, max_y)
        else:
            bins = []

        pathp = self.outdir + '/datavalidation/graphs/'+app+'/actual_data/'
        if not os.path.exists(pathp):
            os.makedirs(pathp)
        
        dv = Datavalidation()
        #l1 = self.make_line_plot(usess_xys, bins, pathp, app, metric, stat, self.dfs_names)
        l2 = dv.make_histograms(dataset_list, bins, pathp, app, metric, self.dfs_names)
        l3 = dv.make_box_cdfs(dataset_list, bins, pathp, app, metric, self.dfs_names)

        tex += self.addfigtex(l2, l3)
        #tex += self.get_tab(dataset_list, bins, pathp, app, metric, self.dfs_names)
        tex += '\n \\clearpage \n'

        return tex

    def plot_selfvalidation(self, app, metric, rep_app_metric_dict):
        mark = ['bo-', 'g^-', 'rx-', 'kd-', 'm+-']

        tex = ''
        
        stats = ['self-KS-2-samp-s']
        for stat in stats:
            all_data = []
            datasets = OrderedDict()
            for name in self.dfs_names:
                if name in rep_app_metric_dict and stat in rep_app_metric_dict[name]:
                    datasets[name] = rep_app_metric_dict[name][stat]
                    all_data.extend(rep_app_metric_dict[name][stat])
                else:
                    datasets[name] = []

            if len(all_data) > 0:
                min_y, max_y = min(all_data), max(all_data)
                bins = get_bins( min_y, max_y)
            else:
                bins = []



            if 'conn_' in metric or '_throughput' in metric:
                print(metric)
                print('%%%%%', len(datasets), datasets.keys(), [len(v) for k,v in datasets.items()])

            pathp = self.outdir + '/selfvalidation/graphs/'+app+'/self_validate/'+stat+'/'
            if not os.path.exists(pathp):
                os.makedirs(pathp)

            sv = Selfvalidation()
            #l1 = sv.make_line_plot(datasets, bins, pathp, app, metric, stat, self.dfs_names)
            l2 = sv.make_histograms(datasets, bins, pathp, app, metric, stat, self.dfs_names)
            l3 = sv.make_box_cdfs(datasets, bins, pathp, app, metric, stat, self.dfs_names)


            tex += self.addfigtex(l2, l3)
            tex += self.get_tab(datasets, bins, pathp, app, metric, self.dfs_names)
            tex += '\n \\clearpage \n'

        return tex

    def plot_crossvalidation(self, app, metric, rep_app_metric_dict):
        
        mark = ['bo-', 'g^-', 'rx-', 'kd-', 'm+-']

        tex = ''

        stats = ['self-KS-2-samp-s']
        for stat in stats:
            #print("curr-stat", stat)
            if 'comparison'  in rep_app_metric_dict:
                #print("rep_app_metric_dict['comparison']", rep_app_metric_dict['comparison'])
                all_data = []
                datasets = OrderedDict()
                comp_names = list(rep_app_metric_dict['comparison'].keys())
                #print('comp_names', comp_names)
                for name in comp_names:
                    #print('name', name)
                    #print('stats', rep_app_metric_dict['comparison'][name].keys())
                    if stat in list(rep_app_metric_dict['comparison'][name].keys()):
                        #print ("rep_app_metric_dict['comparison'][name][stat]", rep_app_metric_dict['comparison'][name][stat])
                        datasets[name] = rep_app_metric_dict['comparison'][name][stat]
                        all_data.extend(rep_app_metric_dict['comparison'][name][stat])
                        #print("datasets", datasets)
                    else:
                        datasets[name] = []
            
                if len(all_data) > 0:
                    min_y, max_y = min(all_data), max(all_data)
                    bins = get_bins( min_y, max_y)
                else:
                    bins = []

                pathp = self.outdir + '/crossvalidation/graphs/'+app+'/cross_validate/'+stat+'/'
                if not os.path.exists(pathp):
                    os.makedirs(pathp)

                #print(datasets)

                cv = Crossvalidation()
                #l1 = cv.make_line_plot(datasets, bins, pathp, app, metric, stat, comp_names)
                l2 = cv.make_histograms(datasets, bins, pathp, app, metric, stat, comp_names)
                l3 = cv.make_box_cdfs(datasets, bins, pathp, app, metric, stat, comp_names)

    
            tex += self.addfigtex(l2, l3)
            tex += self.get_tab(datasets, bins, pathp, app, metric, self.dfs_names)
            tex += '\n \\clearpage \n'

        return tex

    def addfigtex(self, l2, l3):
        tex_tt = ''
        for i in [l2,l3]:
            if i.startswith(self.outdir + '/'):
                i = i[len(self.outdir + '/'):]
            tex_tt +=  '''
                        \\begin{figure}[h!]
                            \\centerline{
                                \\includegraphics[width=1.1\\textwidth,height=0.4\\textheight,keepaspectratio]{%s}
                            }
                        \\end{figure}
                        '''% (i)

        return tex_tt

    def get_tab(self, datasets, bins, pathp, app, metric, pkts_pdus):
        ttx =   '''
                    \\begin{table}[h!]
                    \\centering
                    \\begin{tabular}{p{3.5cm}|p{1.5cm}p{1.5cm}p{1.5cm}p{1.5cm}}
                        \\hline
                             & mean &  median  & SD & Datalen\\\\
                        \\hline
                '''
        
        for i, name in enumerate(pkts_pdus):
            if name in datasets:
                dati = datasets[name]
                me, med, sd, dlen = ['','','','']
                if len(datasets[name]) > 0:
                    me = str(fmt(st.mean(dati)))
                    med = str(fmt(st.median(dati)))
                    sd = str(fmt(st.stdev(dati))) if len(dati) > 1 else 0
                    dlen = str(len(dati))
                ttx += '''
                            %s & %s & %s & %s & %s \\\\
                        ''' %(name, me, med, sd, dlen)

        ttx +=  '''
                    \\hline 
                \\end{tabular}
                \\end{table}

                '''

        return ttx

